#!/bin/bash
#
# Copyright (c) 2012-2014 Fixpoint Inc. All rights reserved.
# ---
# マスターDBからスナップショットを取得し、同期処理を行う
#
shopt -s extglob
THIS_DIR=$(dirname $(readlink -f $0))
SETUP_TYPE="Sync with the Master"

PGSQL_MASTER_SCORE=1000
PGSQL_SLAVE_SCORE=100
RABBITMQ_MASTER_SCORE=10
RABBITMQ_SLAVE_SCORE=5
NA_SCORE="#NA"

. $THIS_DIR/setup_utils.sh

: ${other_name:=$HA_OTHERHOST}
: ${local_name:=}
sync_rabbitmq=false
sync_pgsql=false

usage_exit()
{
    cat <<EOF
Usage: `basename $0` [<master-name>]

  --help                Show this message.

EOF
    exit 1
}

parse_options()
{
    OPTIONS=`getopt -q -o '' -l help -- "$@"`
    [ $? != 0 ] && usage_exit

    local slave_mode
    eval set -- "$OPTIONS"
    while true
    do
        arg="$1"
        case "$arg" in
            --) shift; break ;; # 引数は無視
            *) usage_exit ;;
        esac
        shift
    done

    if [ -n "$1" ]; then
        other_name="$1"
    fi
}

get_rabbitmq_master_score()
{
    crm_attribute -l reboot -N "$local_name" -n master-res_rabbitmq -G -q 2>/dev/null
}

get_pgsql_master_score()
{
    crm_attribute -l reboot -N "$local_name" -n master-res_pgsql -G -q 2>/dev/null
}

get_pgsql_data_status()
{
    crm_attribute -l forever -N "$local_name" -n res_pgsql-data-status -G -q 2>/dev/null
}

get_pgsql_status()
{
    crm_attribute -l reboot -N "$local_name" -n res_pgsql-status -G -q 2>/dev/null
}

get_pgsql_attrset()
{
    echo "$(get_pgsql_status)/$(get_pgsql_data_status)/$(get_pgsql_master_score)"
}

wait_resources_stabilize()
{
    local pgsql_waiting=${1:-true}
    local rabbitmq_waiting=${2:-true}
    local pgsql_stabilized=false
    local rabbitmq_stabilized=false
    local timeout=60
    local stable=10
    local description
    local pgsql_curr pgsql_last pgsql_status pgsql_data_status pgsql_master_score
    local pgsql_count=0
    local pgsql_master="PRI/LATEST/$PGSQL_MASTER_SCORE"
    local pgsql_slave="HS:async/STREAMING|ASYNC/$PGSQL_SLAVE_SCORE"
    local pgsql_pending="HS:alone/LATEST/-INFINITY"
    local rabbitmq_curr rabbitmq_last rabbitmq_master_score
    local rabbitmq_master="$RABBITMQ_MASTER_SCORE"
    local rabbitmq_slave="$RABBITMQ_SLAVE_SCORE"
    local rabbitmq_count=0
    local header="    RabbitMQ  | PostgreSQL                          | description"
    local format="    %-10s| %-36s|%s"

    echo_info "Waiting for the resources to stabilize."
    echo "$header"
    while [ $timeout -gt 0 ]; do
        let timeout=$timeout-1
        #
        # リソースの状態を取得する
        #
        pgsql_status=$(get_pgsql_status)
        pgsql_data_status=$(get_pgsql_data_status)
        pgsql_master_score=$(get_pgsql_master_score)
        if [ -z $pgsql_master_score ]; then
            pgsql_master_score=$NA_SCORE
        fi
        pgsql_curr="$pgsql_status/$pgsql_data_status/$pgsql_master_score"
        rabbitmq_master_score=$(get_rabbitmq_master_score)
        if [ -z $rabbitmq_master_score ]; then
            rabbitmq_master_score=$NA_SCORE
        fi
        rabbitmq_curr="$rabbitmq_master_score"
        description=""
        #
        # pgsql の状態をチェックする
        #
        if $pgsql_waiting; then
            if [[ $pgsql_curr != $pgsql_pending && $pgsql_curr == $pgsql_last ]]; then
                let pgsql_count=$pgsql_count+1
            else
                pgsql_count=0
                pgsql_last=$pgsql_curr
            fi
            if [ $pgsql_curr == $pgsql_master ]; then
                pgsql_stabilized=true
                pgsql_waiting=false
                description="$description PostgreSQL has promoted as a Master."
            elif [ $pgsql_curr == $pgsql_slave ]; then
                pgsql_stabilized=true
                pgsql_waiting=false
                description="$description PostgreSQL has demoted as a Slave."
            elif [ $pgsql_count -ge $stable ]; then
                if [ $pgsql_status == "STOP" ]; then
                    description="$description [WARN] PostgreSQL is stopped."
                else
                    description="$description [WARN] PostgreSQL shows an unexpected state."
                fi
                pgsql_waiting=false
            fi
        fi
        #
        # rabbitmq の状態をチェックする
        #
        if $rabbitmq_waiting; then
            if [[ $rabbitmq_curr == $rabbitmq_last ]]; then
                let rabbitmq_count=$rabbitmq_count+1
            else
                rabbitmq_count=0
                rabbitmq_last=$rabbitmq_curr
            fi
            if [ $rabbitmq_curr == $rabbitmq_master ]; then
                rabbitmq_stabilized=true
                rabbitmq_waiting=false
                description="$description RabbitMQ as promoted as a Master."
            elif [ $rabbitmq_curr == $rabbitmq_slave ]; then
                rabbitmq_stabilized=true
                rabbitmq_waiting=false
                description="$description RabbitMQ as demoted as a Slave."
            elif [ $rabbitmq_count -ge $stable ]; then
                description="$description [WARN] RabbitMQ shows an unexpected state."
                rabbitmq_waiting=false
            fi
        fi
        #
        # 状態を表示する
        #
        printf "$format\n" "$rabbitmq_curr" "$pgsql_curr" "$description"
        #
        # 対象リソースが安定したら終了する
        #
        if $pgsql_stabilized && $rabbitmq_stabilized; then
            echo_info "Resources stablized."
            return 0
        elif ! $pgsql_waiting && ! $rabbitmq_waiting; then
            break
        fi
        sleep 1
    done
    return 1
}

check_standby()
{
    #
    # スタンバイモードの確認
    #
    if [ "$(crm_standby -qG)" == "on" ]; then
        abort_setup "This node is currently in standby mode. Please execute 'pcs cluster unstandby' and try again."
    fi
}

check_nodename()
{
    #
    # ノード名の確認
    #
    if [ -z $local_name ]; then
        local_name=$(crm_node -n)
        if [ -z $local_name ]; then
            abort_setup "Name of local node could not be determined."
        fi
    fi
    if [ -z $other_name ]; then
        other_name=$(crm_node -l | cut -d' ' -f2 | grep -v -w $(crm_node -n) | head -n 1)
        if [ -z $other_name ]; then
            abort_setup "Name of other node could not be determined."
        fi
    fi
    echo_info "local_name=$local_name"
    echo_info "other_name=$other_name"
}

check_pacemaker()
{
    check_standby
    check_nodename
}

start_pacemaker()
{
    echo_title "Start pacemaker service."
    #
    # pacemaker 制御下でなく起動している pgsql/rabbitmq は停止させておく
    #
    if is_activce_service $POSTGRES; then
        echo_warn "Stop $POSTGRES not under pacemaker control."
        service_stop $POSTGRES
    fi
    if is_activce_service rabbitmq-server; then
        echo_warn "Stop rabbitmq-server not under pacemaker control."
        killall -q beam beam.smp
        service_stop rabbitmq-server
    fi
    #
    # pacemaker を起動する
    #
    service_start pacemaker
    exit_if_failed "$?" "Failed to start pacemaker."
    cluster_wait_current_dc
    exit_if_failed "$?" "The state of pacemaker did not stabilize."
    echo_info "Pacemaker has stablized."
    check_pacemaker
    #
    # リソースの状態が安定するのを待つ
    #
    wait_resources_stabilize
}

check_cluster()
{
    echo_title "Check cluster status."
    get_cluster_status
    if ! $CLUSTER_CONFIGURED; then
        abort_setup "Cluster configuration is not setup."
    fi
    if ! $CLUSTER_RUNNING; then
        # pacemaker が起動していないときは起動させる
        start_pacemaker
    else
        check_pacemaker
    fi
}

check_pgsql()
{
    #
    # マスターノードでは sync_master しない
    #
    if pgsql_is_primary; then
        echo_wan "PostgreSQL on this node is running as a Master."
        return
    fi
    #
    # pacemaker が管理している postgres の状態をチェックする
    # pgsql_status が STOP でなく pgsql_master_score が 100 のときは
    # 同期状態でレプリケーション中とみなす（sync_master 不要）
    #
    local pgsql_status=$(get_pgsql_status)
    local pgsql_master_score=$(get_pgsql_master_score)
    if [[ -n $pgsql_status && $pgsql_status != "STOP" && $pgsql_master_score == $PGSQL_SLAVE_SCORE ]]; then
        # TODO: --force オプションで強制的に sync_master させる
        echo_info "PostgreSQL on this node is running as a Slave."
        return
    fi
    echo_info "PostgreSQL on this node needs to be synchronized with the Master."
    sync_pgsql=true
}

check_rabbitmq()
{
    #
    # rabbitmq がスレーブとして動作中かチェックする
    #
    local rabbitmq_master_score=$(get_rabbitmq_master_score)
    if [[ $rabbitmq_master_score == $RABBITMQ_MASTER_SCORE ]]; then
        echo_warn "RabbitMQ on this node is running as a Master."
        return
    elif [[ $rabbitmq_master_score == $RABBITMQ_SLAVE_SCORE ]]; then
        # TODO: --force オプションで強制的に sync_master させる
        echo_info "RabbitMQ on this node is running as a Slave."
        return
    fi
    echo_info "RabbitMQ on this node needs to be synchronized with the Master."
    sync_rabbitmq=true
}

sync_rabbitmq()
{
    #
    # rabbitmq の同期
    #
    if $sync_rabbitmq; then
        #
        # rabbitmq-server が動作中なら停止させる
        #
        if is_activce_service rabbitmq-server; then
            killall -q beam beam.smp
            if ! service_stop rabbitmq-server; then
                echo_error "Failed to stop the RabbitMQ."
                return 1
            fi
        fi
    fi
    return 0
}

sync_pgsql()
{
    #
    # pgsql の復旧
    #
    if $sync_pgsql; then
        #
        # postgres が動作中なら停止させる
        #
        if is_activce_service $POSTGRES; then
            service_stop $POSTGRES
        fi
        #
        # データベースのバックアップを実行する
        #
        if ! pgsql_replica $other_name; then
            echo_error "Failed to backup the database."
            return 1
        fi
        #
        # データベース同期後の kompirad 起動時は DB マイグレーション不要
        # になるため、$KOMPIRA_VAR_DIR/startup/do* ファイルを削除しておく
        #
        rm -f $KOMPIRA_VAR_DIR/startup/{do_migrate,do_import_data,do_compile_jobflow}
    fi
    return 0
}

sync_master()
{
    local rc_pgsql rc_rabbitmq
    local rc=0
    #
    # メンテナンスモードに移行する
    #
    pcs_enter_maintenance_mode
    exit_if_failed "$?" "Failed to enter maintenance mode."
    #
    # pgsql を同期させる
    #
    sync_pgsql
    rc_pgsql=$?
    #
    # rabbitmq を同期させる
    #
    sync_rabbitmq
    rc_rabbitmq=$?
    #
    # 同期できれば失敗カウンタをリセットする
    #
    if [ $rc_rabbitmq == 0 ]; then
        echo_info "Reset the failcount of RabbitMQ."
        pcs_failcount_reset res_rabbitmq

    else
        rc=1
    fi
    if [ $rc_pgsql == 0 ]; then
        echo_info "Reset the failcount of PostgreSQL"
        pcs_failcount_reset res_pgsql
    else
        rc=1
    fi
    if [ $rc == 0 ]; then
        pcs_failcount_reset res_vip res_httpd res_kompirad res_kompira_jobmngrd res_lsyncd
    fi
    #
    # メンテナンスモードを解除する
    #
    pcs_leave_maintenance_mode
    #
    # 同期の成否をチェックする
    #
    exit_if_failed "$rc" "Failed to sync with $other_name."
}

################################################################

if ! is_yum_installed "pacemaker"; then
    echo_error "Cluster configuration is not setup."
    exit 1
fi
parse_options "$@"
start_setup
check_cluster
check_pgsql
check_rabbitmq
if $sync_pgsql || $sync_rabbitmq; then
    sync_master
    wait_resources_stabilize
    exit_if_failed $? "Resources did not stabilize."
fi
pcs_show_status
exit_setup 0
