#!/bin/bash
#
# Copyright (c) 2012-2014 Fixpoint Inc. All rights reserved.
# ---
# Kompiraクラスタ構成セットアップスクリプト
#
# * NICは以下のように割当てる::
#   eth0 --- サービス提供用インタフェース
#   eth1 --- heartbeat用インタフェース
#
# * 各パラメータは以下のとおりとする
#                             primary         secondary
#   heartbeat ipaddress       192.168.99.1    192.168.99.2
#   hostname                  kompira-server1 kompira-server2
#
# * PostgreSQLのpostgresパスワードの設定は同じにしておくこと
#
shopt -s extglob
SETUP_TYPE="Setup the Kompira Cluster"
SETUP_LOG="setup_cluster.$$.log"
THIS_DIR=$(dirname $(readlink -f $0))

. $THIS_DIR/scripts/setup_utils.sh

HEARTBEAT_DEVICE=
HEARTBEAT_NETCIDR=192.168.99.0/24
HEARTBEAT_UNICAST=false
CLUSTER_VIP=
CLUSTER_DEVICE=
PRIMARY_IP=
SECONDARY_IP=
# YUMOPT_FOR_HA=

CMAN_MODE=false
PCSD_MODE=true
SECONDARY=false
SLAVE_MODE=false
SKIP_HOSTNAME=true
SKIP_HOSTS=false
SKIP_HBNET=false
SKIP_FIREWALL=false
SKIP_IPTABLES=false
MANUAL_HEARTBEAT=false
NO_JOBMNGR_SETUP=false
NO_CLUSTER_VIP=false
MANUAL_MODE=false
DRY_RUN=false

HACLUSTER_USER=hacluster
HACLUSTER_PASS=hacluster

#
# RHEL High-Availability リポジトリ情報
#
RHEL_HA_REPONAME="rhel-ha-for-rhel-${RHEL_VERSION}-server-rpms rhel-ha-for-rhel-${RHEL_VERSION}-server-eus-rpms"
PACKAGES_FOR_HA_EL6="pacemaker pcs xinetd rsync lsyncd patch"
PACKAGES_FOR_HA_EL7="pcs rsync lsyncd patch"

#
# 冗長構成に必要なパッケージ
#
if [ $RHEL_VERSION == 6 ]; then
    CMAN_MODE=true
    PACKAGES_FOR_HA=$PACKAGES_FOR_HA_EL6
elif [ $RHEL_VERSION == 7 ]; then
    PACKAGES_FOR_HA=$PACKAGES_FOR_HA_EL7
else
    abort_setup "Unexpected RHEL version: $RHEL_VERSION"
fi

#
# RHEL では High-Availability リポジトリを有効にする
#
#if [ $SYSTEM == "RHEL" ]; then
#    for reponame in $RHEL_HA_REPONAME; do
#        YUMOPT_FOR_HA="${RHEL_HA_REPONAME[@]/#/--enablerepo=}"
#    done
#fi

#
# SYSTEMD かどうかでリソースエージェントを切り分ける
#
if $SYSTEMD; then
    PCS_STD_RESOURCE_CLASS=systemd
else
    PCS_STD_RESOURCE_CLASS=lsb
fi

: ${ETH0:="eth0"}
: ${ETH1:="eth1"}

: ${HA_CLUSTER_NAME:=""}
: ${HA_HOSTNAME_PREFIX:="ha-kompira"}
: ${PRIMARY_SUFFIX:="1"}
: ${SECONDARY_SUFFIX:="2"}
: ${PRIMARY_IP_OFFSET:=1}
: ${SECONDARY_IP_OFFSET:=2}

: ${cluster_name:=}
: ${primary_name:=}
: ${secondary_name:=}
: ${primary_altname:=}
: ${secondary_altname:=}
: ${primary_ip:=}
: ${secondary_ip:=}
: ${local_hostname:=}
: ${other_hostname:=}
: ${local_ipaddr:=}
: ${other_ipaddr:=}

: ${heartbeat_baseaddr:=}
: ${heartbeat_ipaddr:=}
: ${cluster_nic:=}
: ${cluster_vip:=}
: ${cluster_netmask:=}

SHOW_OPTIONS="$SHOW_OPTIONS SECONDARY SLAVE_MODE NO_JOBMNGR_SETUP HA_HOSTNAME_PREFIX MANUAL_HEARTBEAT PRIMARY_IP SECONDARY_IP HEARTBEAT_NETCIDR CLUSTER_VIP DRY_RUN"
SHOW_PARAMS="cluster_name primary_name secondary_name heartbeat_baseaddr heartbeat_ipaddr cluster_nic cluster_vip cluster_netmask primary_ip secondary_ip local_ipaddr"

ipaddr2int()
{
    python -c "import socket; import struct; print struct.unpack('>I', socket.inet_aton('$1'))[0]"
}

int2ipaddr()
{
    python -c "import socket; import struct; print socket.inet_ntoa(struct.pack('>I', $1))"
}

check_ipaddr()
{
    ipcalc -c $1
}

check_cidr()
{
    local cidr=$1
    echo $cidr | egrep '^([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})/([0-9]{1,2})' >/dev/null && ipcalc -c $cidr
}

check_ipaddr_in_netcidr()
{
    local label=$1
    local ipaddr=$2
    local netcidr=$3
    local netmask=${netcidr#*/}
    #
    # ipaddr が netcidr の範囲に含まれていること、および
    # ネットワークまたはブロードキャストアドレスになっていないことをチェックする
    #
    if [ "$(ipcalc -n $ipaddr/$netmask)" != "$(ipcalc -n $netcidr)" ]; then
        echo_error "$label($ipaddr) not in $netcidr"
        return 1
    fi
    if [ "$(ipcalc -n $ipaddr/$netmask)" == "NETWORK=$ipaddr" ]; then
        echo_error "$label($ipaddr) is network address"
        return 1
    fi
    if [ "$(ipcalc -b $ipaddr/$netmask)" == "BROADCAST=$ipaddr" ]; then
        echo_error "$label($ipaddr) is broadcast address"
        return 1
    fi
    return 0
}

show_params()
{
    echo_info ""
    show_options $SHOW_PARAMS
    echo_info ""
}

set_params()
{
    echo_title "Preparing the parameters."

    if [ -n "$PRIMARY_NAME" ]; then
        primary_name="$PRIMARY_NAME"
    else
        primary_name="${HA_HOSTNAME_PREFIX}${PRIMARY_SUFFIX}"
    fi
    if [ -n "$SECONDARY_NAME" ]; then
        secondary_name="$SECONDARY_NAME"
    else
        secondary_name="${HA_HOSTNAME_PREFIX}${SECONDARY_SUFFIX}"
    fi

    #
    # ハートビート用 CIDR から、プライマリ／セカンダリ用 IP アドレスを計算する
    #
    if $MANUAL_HEARTBEAT; then
        primary_ip=$PRIMARY_IP
        secondary_ip=$SECONDARY_IP
        HEARTBEAT_UNICAST=true
    else
        heartbeat_baseaddr=${HEARTBEAT_NETCIDR%/*}
        primary_ip=$(int2ipaddr $(($(ipaddr2int $heartbeat_baseaddr)+$PRIMARY_IP_OFFSET)))
        secondary_ip=$(int2ipaddr $(($(ipaddr2int $heartbeat_baseaddr)+$SECONDARY_IP_OFFSET)))
    fi
    if ! $NO_CLUSTER_VIP; then
        cluster_vip=${CLUSTER_VIP%/*}
        cluster_netmask=${CLUSTER_VIP#*/}
    fi
    if [ -n "$CLUSTER_DEVICE" ]; then
        cluster_nic="$CLUSTER_DEVICE"
    fi
    #
    # クラスタ名称を生成する
    #
    cluster_name="${HA_CLUSTER_NAME}"
    if [ -z "$cluster_name" ]; then
        local cluster_id
        if ! $MANUAL_HEARTBEAT; then
            cluster_id="${HA_HOSTNAME_PREFIX}${HEARTBEAT_NETCIDR}"
        else
            cluster_id="${HA_HOSTNAME_PREFIX}${primary_ip}${secondary_ip}"
        fi
        if $CMAN_MODE; then
            # [CENT6/RHEL6] CMAN ではクラスタ名は最大15文字(cman 3.0)
            cluster_name="ha-kompira-$(echo "$cluster_id" | sha1sum | cut -c1-4)"
        else
            cluster_name="ha-kompira-$(echo "$cluster_id" | sha1sum | cut -c1-6)"
        fi
    fi

    if ! $SECONDARY; then
        local_hostname=$primary_name
        other_hostname=$secondary_name
        local_ipaddr=$primary_ip
        other_ipaddr=$secondary_ip
        heartbeat_ipaddr=$primary_ip
        primary_altname=$HA_LOCALHOST
        secondary_altname=$HA_OTHERHOST
    else
        local_hostname=$secondary_name
        other_hostname=$primary_name
        local_ipaddr=$secondary_ip
        other_ipaddr=$primary_ip
        heartbeat_ipaddr=$secondary_ip
        primary_altname=$HA_OTHERHOST
        secondary_altname=$HA_LOCALHOST
    fi
}

check_params()
{
    if ! $MANUAL_HEARTBEAT; then
        echo_title "Check the paramters"
        #
        # 各アドレスが HEARTBEAT_NETCIDR の範囲にあるかをチェック
        #
        local result=0
        check_ipaddr_in_netcidr "heartbeat_ipaddr" $heartbeat_ipaddr $HEARTBEAT_NETCIDR
        let result+=$?
        check_ipaddr_in_netcidr "primary_ip" $primary_ip $HEARTBEAT_NETCIDR
        let result+=$?
        check_ipaddr_in_netcidr "secondary_ip" $secondary_ip $HEARTBEAT_NETCIDR
        let result+=$?
        exit_if_failed $result "invalid heartbeat-netaddr ($HEARTBEAT_NETCIDR)"
    fi
}

check_environ()
{
    echo_title "Check the current environment"
    if [ -n "$PROXY_URL" ]; then
        echo_debug "settings http_proxy for install: $PROXY_URL"
        echo_debug "settings no_proxy for install: $NO_PROXY"
        export http_proxy="$PROXY_URL"
        export https_proxy="$PROXY_URL"
        export HTTP_PROXY="$PROXY_URL"
        export HTTPS_PROXY="$PROXY_URL"
        export no_proxy="$NO_PROXY"
    fi

    check_arch
    check_system
    check_subscription

    #
    # RHEL High Availability アドオンのチェック
    #
    if [ $SYSTEM == "RHEL" ]; then
        if ! $OFFLINE_MODE; then
            verbose_run yum repolist --enablerepo="${RHEL_HA_REPONAME// /,}"
            exit_if_failed "$?" "Addon 'RHEL High Availability' is not available"
        fi
    fi

    if $CMAN_MODE; then
        #
        # [CENT6/RHEL6] CMAN では NetworkManager が起動していないこと
        #
        local chkconfig_nm
        chkconfig_nm=$(chkconfig --list NetworkManager 2> /dev/null)
        if [ $? == 0 ]; then
            if echo "$chkconfig_nm" | grep ":on"; then
                abort_setup "NetworkManager is enabled"
            fi
            if $SERVICE NetworkManager status; then
                abort_setup "NetworkManager is running"
            fi
        fi
    fi
    if [ -n "$HEARTBEAT_DEVICE" ]; then
        #
        # HEARTBEAT_DEVICE が指定されている場合、デバイスが存在することを確認する
        #
        ip link show $HEARTBEAT_DEVICE > /dev/null
        exit_if_failed "$?" "Device $HEARTBEAT_DEVICE does not exist!"
        #
        # ハートビートIPアドレスが設定されているかチェックする
        #
        if ip -4 -o addr show scope global dev $HEARTBEAT_DEVICE | grep -w -q $local_ipaddr; then
            # 指定したデバイスに既に設定済みなのでなにもしない
            echo_info "Address $local_ipaddr configured on $HEARTBEAT_DEVICE already."
            SKIP_HBNET=true
        elif ip -4 -o addr show scope global | grep -w -q $local_ipaddr; then
            # 指定したデバイス以外で既にアドレスが設定されている場合はエラー
            abort_setup "Address $local_ipaddr configured on other device!"
        fi
    elif $MANUAL_HEARTBEAT; then
        #
        # HEARTBEAT_DEVICE が指定されておらず、ハートビート手動設定の場合は、
        # いずれかのデバイスに $local_ipaddr が設定済みであることを確認する
        #
        if ! ip -4 -o addr show scope global | grep -w -q $local_ipaddr; then
            abort_setup "$?" "Address $local_ipaddr does not configured!"
        fi
        SKIP_HBNET=true
    fi

    #
    # Kompira 用 virtualenv が設定されていること [1.4.11]
    #
    if [ ! -x $KOMPIRA_ENV/bin/pip ]; then
        abort_setup "Kompira virtualenv ($KOMPIRA_ENV) is not configured"
    fi
    local -x PIP=$KOMPIRA_ENV/bin/pip

    #
    # Kompira がインストールされていること／1.4以上であることを確認する
    #
    local kompira_version=$(get_kompira_version)
    case $kompira_version in
        "")
            abort_setup "Kompira is not installed"
            ;;
        "1.[^0-3].*")
            abort_setup "Kompira version 1.4 or higher is required! ($kompira_version)"
            ;;
        *)
            echo_info "Confirmed Kompira version: $kompira_version"
            ;;
    esac
}

install_cluster_packages()
{
    echo_title "Install RPM packages for High Availability"
    local yum_options
    if ! $OFFLINE_MODE; then
        # yum_options=$YUMOPT_FOR_HA
        verbose_run yum-config-manager --enable $RHEL_HA_REPONAME
    fi

    verbose_run yum clean all
    YUM_OPTION=$yum_options yum_install $PACKAGES_FOR_HA
}

setup_nic_nmcli()
{
    local dev=$1
    local ipaddr=$2

    echo_info "Configure NIC $dev settings [nmcli]"
    if ! LANG= nmcli connection show $dev >/dev/null 2>&1; then
       LANG= nmcli connection add type ethernet ifname $dev con-name $dev
    fi
    LANG= nmcli connection modify $dev ipv4.method manual +ipv4.addresses $ipaddr
    echo_info "Restart the NIC $dev"
    LANG= nmcli connection up $dev
    exit_if_failed "$?" "Failed to restart NIC $dev"
}

setup_nic_ifcfg()
{
    local dev=$1
    local ipaddr=$2
    local ifcfg_file=$NETWORK_SCRIPTS_DIR/ifcfg-$dev

    echo_info "Configure NIC $dev settings [ifcfg]"
    if [ -f $ifcfg_file ]; then
        cp $ifcfg_file $TMPDIR/ifcfg-$dev.tmp
    else
        echo -e "DEVICE=$dev\nONBOOT=no\nBOOTPROTO=none\nNM_CONTROLLED=no" > $TMPDIR/ifcfg-$dev.tmp
    fi
    ( sed -e "/DEVICE=/c DEVICE=$dev" \
          -e "/ONBOOT=/c ONBOOT=yes" \
          -e "/BOOTPROTO=/c BOOTPROTO=none" \
          -e "/NM_CONTROLLED=/c NM_CONTROLLED=no" \
          -e "/DHCP_HOSTNAME=/d" \
          -e "/HOSTNAME=/d" \
          $TMPDIR/ifcfg-$dev.tmp
    ) > $TMPDIR/ifcfg-$dev

    # IPADDR(\d)*= という行で \d 部分が最大のものを抜き出して、追加するインデックス番号を決める
    local INDEX=""
    local MAXIPADDR=$(grep -E "^IPADDR[0-9]*=" $TMPDIR/ifcfg-$dev | cut -d= -f1 | sort -k 1.7n | tail -n 1)
    if [ -n "$MAXIPADDR" ]; then
        INDEX=$((${MAXIPADDR#IPADDR}+1))
    fi
    ( echo "IPADDR$INDEX=${ipaddr%/*}"
      echo "PREFIX$INDEX=${ipaddr#*/}"
    ) >> $TMPDIR/ifcfg-$dev
    # exit_if_failed "$?" "Failed to config $dev"

    # [注意] .old を suffix にすると $eth.old というインタフェースが作られてしまう
    SIMPLE_BACKUP_SUFFIX='~' diff_cp $TMPDIR/ifcfg-$dev $ifcfg_file

    echo_info "Restart the NIC $dev"
    ifdown $dev
    ifup $dev
    exit_if_failed "$?" "Failed to restart NIC $dev"
}

setup_heatbeat_network()
{
    local dev=$HEARTBEAT_DEVICE
    local ipaddr="$heartbeat_ipaddr/${HEARTBEAT_NETCIDR#*/}"
    if [ $RHEL_VERSION == 6 ]; then
        setup_nic_ifcfg $dev $ipaddr
    elif [ $RHEL_VERSION == 7 ]; then
        setup_nic_nmcli $dev $ipaddr
    fi
}

setup_hostname()
{
    echo_info "Configure hostname: $local_hostname"

    local network_file=/etc/sysconfig/network
    if grep -q "HOSTNAME" $network_file; then
        sed -e "/HOSTNAME\b/c HOSTNAME=$local_hostname" $network_file > $TMPDIR/network
    else
        (cat $network_file; echo "HOSTNAME=$local_hostname") > $TMPDIR/network
    fi
    diff_cp $TMPDIR/network $network_file
    hostname $local_hostname
    HOSTNAME=$local_hostname
}

setup_hosts()
{
    echo_info "Configure /etc/hosts settings"

    local hosts_file=/etc/hosts
    #
    # primary_name, secondary_name の記述があれば削除して
    # primary_ip, secondary_ip の行は追加しなおす
    #
    verbose_run sed -r \
                -e "s/\s+($primary_name|$secondary_name|$HA_LOCALHOST|$HA_OTHERHOST)\b/ /g" \
                -e "s/\s+$//" \
                -e "/^[0-9.]+$/d" \
                $hosts_file > $TMPDIR/hosts
    echo "$primary_ip $primary_name $primary_altname" >> $TMPDIR/hosts
    echo "$secondary_ip $secondary_name $secondary_altname" >> $TMPDIR/hosts
    diff_cp $TMPDIR/hosts $hosts_file
}

setup_network()
{
    echo_title "Configure Network settings"

    if ! $SKIP_HBNET; then
        setup_heatbeat_network
    fi
    if ! $SKIP_FIREWALL; then
        if is_activce_service firewalld; then
            # "2224/tcp" "4369/tcp" "5672/tcp" "15672/tcp" "25672/tcp"
            setup_firewalld "2224/tcp" "$other_ipaddr/32"
        elif is_activce_service iptables; then
            setup_iptables "$other_ipaddr/32"
        fi
    fi
    if ! $SKIP_HOSTNAME; then
        setup_hostname
    fi
    if ! $SKIP_HOSTS; then
        setup_hosts
    fi
}

setup_apache()
{
    echo_title "Configure the Apache service."
    diff_cp $THIS_DIR/config/cluster/status.conf /etc/httpd/conf.d/status.conf

    echo_info "Disable and stop the Apache service."
    service_disable httpd
    service_stop httpd
}

setup_pgsql()
{
    echo_title "Configure the PostgreSQL service"

    verbose_run $INSTALL -o postgres -g postgres -m 700 -d "$PG_BASEDIR/pg_archive"

    # $CLUSTER_RESOURCE_PATH/heartbeat/pgsql スクリプトにある su を runuser に置き換える
    verbose_run sed -re "s|\b(su)\b|$SU|" $CLUSTER_RESOURCE_PATH/heartbeat/pgsql > $TMPDIR/pgsql
    diff_install -m 755 $TMPDIR/pgsql $CLUSTER_RESOURCE_PATH/heartbeat/pgsql

    echo_info "Disable and stop the PostgreSQL service"
    service_disable $PG_SERVICE
    service_stop $PG_SERVICE

    if ! $SLAVE_MODE; then
       echo_title "Configure PostgreSQL streaming replication settings."
       verbose_run sed -re "/^host replication $PG_REPL_USER/d" $PG_DATADIR/pg_hba.conf > $TMPDIR/pg_hba.conf
       echo "host replication $PG_REPL_USER $primary_name trust" >> $TMPDIR/pg_hba.conf
       echo "host replication $PG_REPL_USER $secondary_name trust" >> $TMPDIR/pg_hba.conf
       diff_install -o postgres -g postgres -m 600 $TMPDIR/pg_hba.conf $PG_DATADIR/pg_hba.conf
       diff_install -o postgres -g postgres -m 600 $THIS_DIR/config/cluster/postgresql.conf $PG_DATADIR/postgresql.conf
       verbose_run rm -f $PG_BASEDIR/tmp/PGSQL.lock

       echo_info "Create PostgreSQL replication user: $PG_REPL_USER"
       service_start $PG_SERVICE
       verbose_run psql -U $PGSQL_USER -c "create user $PG_REPL_USER password '$PG_REPL_PASS' NOSUPERUSER REPLICATION"
       service_stop $PG_SERVICE
    fi
}

setup_rabbitmq()
{
    echo_title "Configure RabbitMQ clustering settings."

    echo_info "Disable the RabbitMQ service"
    service_disable rabbitmq-server
    service_stop rabbitmq-server

    echo_info "Configure rabbitmq cluster configs."
    echo -n "$cluster_name" > $TMPDIR/dot.erlang.cookie
    diff_install -o rabbitmq -g rabbitmq -m 400 $TMPDIR/dot.erlang.cookie /var/lib/rabbitmq/.erlang.cookie
    expand_template $THIS_DIR/config/cluster/rabbitmq.config > $TMPDIR/rabbitmq.config
    diff_cp $TMPDIR/rabbitmq.config /etc/rabbitmq/rabbitmq.config
    echo "NODENAME=rabbit@$local_hostname" > $TMPDIR/rabbitmq-env.conf
    diff_cp $TMPDIR/rabbitmq-env.conf /etc/rabbitmq/rabbitmq-env.conf

    # /usr/sbin/rabbitmq* スクリプトにある su を runuser に置き換える
    local rmq_sbin
    for rmq_sbin in rabbitmq-plugins rabbitmq-server rabbitmqctl; do
        if [ -f /usr/sbin/$rmq_sbin ]; then
            verbose_run sed -re "s|\b(su)\b|$SU|" /usr/sbin/$rmq_sbin > $TMPDIR/$rmq_sbin
            diff_install -m 755 $TMPDIR/$rmq_sbin /usr/sbin/$rmq_sbin
        fi
    done

    if ! $SLAVE_MODE; then
        local rpm_ver=$(get_rpm_version rabbitmq-server)

        service_start rabbitmq-server
        if [ $(ver2num $rpm_ver) -ge $(ver2num "3.3.0") ]; then
            # 古い rabbitmqctl ではクラスタ名の設定はできない(詳細未確認だが仮に 3.3.0 未満で判断)
            verbose_run rabbitmqctl set_cluster_name "$cluster_name"
        fi
        verbose_run rabbitmqctl set_policy all '.*' '{"ha-mode": "all"}'
        service_stop rabbitmq-server
    fi
}

setup_rsync()
{
    echo_title "Configure rsyncd/lsyncd settings for directory syncing."

    diff_cp $THIS_DIR/config/cluster/rsyncd.conf /etc/rsyncd.conf
    if [ $RHEL_VERSION == 6 ]; then
        verbose_run sed -e "s/disable\s*=\s*yes/disable = no/" /etc/xinetd.d/rsync > $TMPDIR/rsync
        SIMPLE_BACKUP_SUFFIX='~' diff_cp $TMPDIR/rsync /etc/xinetd.d/rsync
        service_restart xinetd
        service_enable xinetd
    elif [ $RHEL_VERSION == 7 ]; then
        service_restart rsyncd
        service_enable rsyncd
    fi

    verbose_run sed -e 's/^LSYNCD_OPTIONS=" "/#LSYNCD_OPTIONS=" "/' /etc/sysconfig/lsyncd > $TMPDIR/lsyncd
    diff_cp $TMPDIR/lsyncd /etc/sysconfig/lsyncd
    expand_template $THIS_DIR/config/cluster/lsyncd.conf > $TMPDIR/lsyncd.conf
    diff_cp $TMPDIR/lsyncd.conf /etc/lsyncd.conf
    service_disable lsyncd
    service_stop lsyncd
}

setup_kompira()
{
    echo_title "Disable and stop the Kompira-Daemon service"

    service_disable kompirad
    service_disable kompira_jobmngrd
    service_stop kompirad
    service_stop kompira_jobmngrd

    #
    # systemd では Requires 設定を除外する（依存関係は pacemaker 側で制御する）
    # kompira をアップデートすると /usr/lib/systemd/system/kompira*.service を更新するが、
    # その影響を受けないようにするために /etc/systemd/system に配置する
    #
    if $SYSTEMD; then
        local svcname
        for svcname in kompirad.service kompira_jobmngrd.service; do
            verbose_run sed -re "s/^\s*(Requires\s*=.*)/# \1/" /usr/lib/systemd/system/$svcname > $TMPDIR/$svcname
            diff_cp $TMPDIR/$svcname /etc/systemd/system/$svcname
        done
        verbose_run $SYSTEMCTL daemon-reload
    fi

    # スクリプトを更新する
    diff_install -m 644 scripts/setup_utils.sh $KOMPIRA_BIN/setup_utils.sh
    diff_install -m 755 scripts/sync_master.sh $KOMPIRA_BIN/sync_master.sh
}

cluster_restart()
{
    if $PCSD_MODE; then
        cluster_stop
        cluster_start
    else
        service_restart pacemaker
    fi
}

cluster_start()
{
    if $PCSD_MODE; then
        verbose_run pcs cluster start
    else
        service_start pacemaker
    fi
}

cluster_stop()
{
    if $PCSD_MODE; then
        verbose_run pcs cluster stop
    else
        service_stop pacemaker
    fi
}

setup_pcsd()
{
    echo_title "Configure pcsd settings"
    local pcs_options
    if $HEARTBEAT_UNICAST; then
        pcs_options="--transport udpu"
    fi
    echo "$HACLUSTER_PASS" | LANG= verbose_run passwd $HACLUSTER_USER --stdin
    service_enable pcsd
    service_start pcsd
    if ! $SECONDARY; then
        #
        # まずはプライマリノードでのみ pcs の認証を実施する
        #
        verbose_run pcs cluster auth --local --force $primary_name -u $HACLUSTER_USER -p "$HACLUSTER_PASS"
    else
        verbose_run pcs cluster auth --force $primary_name $secondary_name -u $HACLUSTER_USER -p "$HACLUSTER_PASS"
    fi
    verbose_run pcs cluster setup --local --enable --name "$cluster_name" $primary_name $secondary_name $pcs_options
}

setup_cman()
{
    echo_title "Configure cman settings"

    local cman_options
    if $HEARTBEAT_UNICAST; then
        cman_options='transport="udpu"'
    fi
    diff_cp $THIS_DIR/config/cluster/cman /etc/sysconfig/cman
    sed \
        -e "s/<KOMPIRA-CLUSTER-NAME>/$cluster_name/" \
        -e "s/<KOMPIRA-PRIMARY-NAME>/$primary_name/" \
        -e "s/<KOMPIRA-SECONDARY-NAME>/$secondary_name/" \
        -e "s/<CMAN-OPTIONS>/$cman_options/" \
        $THIS_DIR/config/cluster/cluster.conf > $TMPDIR/cluster.conf
    verbose_run ccs_config_validate -l $TMPDIR/cluster.conf
    exit_if_failed "$?" "Invalid cluster.conf"
    diff_cp $TMPDIR/cluster.conf /etc/cluster/cluster.conf
    diff_cp $THIS_DIR/config/cluster/logrotate.corosync /etc/logrotate.d/corosync
}

stop_pacemaker()
{
    ## if $CLUSTER_RUNNING; then
    echo_title "Stop the Pacemaker"
    cluster_stop
    ## fi
}

patch_resource_agent()
{
    local rpm_ver=$(get_rpm_version resource-agents)
    echo_info "Patch resource-agents $rpm_ver"
    #
    # apache-conf.sh の source_envfiles で読み取った変数を
    # httpd に環境変数として渡すために export するようにパッチをあてる
    #
    if patch --dry-run -f -s -d/ -p1 -i $THIS_DIR/patches/apache-conf.sh.patch > /dev/null; then
        verbose_run patch -b -f -d/ -p1 -i $THIS_DIR/patches/apache-conf.sh.patch
    fi
    #
    # resource-agents 3.9.6 未満の場合
    # /usr/lib/ocf/lib/heartbeat/ocf-shellfuncs の ocf_local_nodename が
    # 適切に crm_node -n でノード名を扱うようにパッチをあてる
    #
    # https://github.com/ClusterLabs/resource-agents/commit/5adecc2af9b0a2e67fd70629a0a68714474879b0
    #
    if [ $(ver2num $rpm_ver) -lt $(ver2num "3.9.6") ]; then
        if patch --dry-run -f -s -d/ -p1 -i $THIS_DIR/patches/ocf-shellfuncs.patch > /dev/null; then
            verbose_run patch -b -f -d/ -p1 -i $THIS_DIR/patches/ocf-shellfuncs.patch
        fi
    fi
}

parch_pcs()
{
    local rpm_ver=$(get_rpm_version pcs)
    echo_info "Patch pcs $rpm_ver: $pcs_path"
    #
    # pcs 0.9.160 未満の場合
    # pcs status コマンドなどで UnicodeDecodeError が発生するためパッチをあてる
    #
    # https://github.com/ClusterLabs/pcs/commit/52ea51c0006e585345199e4a3b3a8565e741d024
    #
    if [ $(ver2num $rpm_ver) -lt $(ver2num "0.9.160") ]; then
        local pcs_path=$(rpm -q pcs -l | grep "/site-packages/pcs$" | head -1)
        if [ -z $pcs_path ]; then
            return
        fi
        verbose_run patch -b -f -d$pcs_path -p2 -i $THIS_DIR/patches/pcs-remove_import_unicode_literals.patch | grep "^patching file"
    fi
}

setup_pacemaker()
{
    if $PCSD_MODE; then
        setup_pcsd
    else
        setup_cman
    fi
    patch_resource_agent
    parch_pcs
    #
    # kompirad/kompirad_jobmngrd/rabbitmq-server 用のリソースファイルをコピー
    # デフォルトで入る rabbitmq-server スクリプトは promote/demote に
    # 対応していないため独自のスクリプトに置き換える
    #
    echo_info "Copy resource script files to $CLUSTER_RESOURCE_PATH"
    verbose_run $INSTALL -d $CLUSTER_RESOURCE_PATH/kompira
    verbose_run $INSTALL -m 755 -S .old -b $THIS_DIR/config/cluster/kompirad $CLUSTER_RESOURCE_PATH/kompira/kompirad
    verbose_run $INSTALL -m 755 -S .old -b $THIS_DIR/config/cluster/kompira_jobmngrd $CLUSTER_RESOURCE_PATH/kompira/kompira_jobmngrd
    verbose_run $INSTALL -m 755 -S .old -b $THIS_DIR/config/cluster/rabbitmq-server $CLUSTER_RESOURCE_PATH/rabbitmq/rabbitmq-server

    echo_info "Enable the Pacemaker service"
    service_enable pacemaker
}

start_slave()
{
    echo_title "Start the Pacemaker as SLAVE"
    pgsql_replica $HA_OTHERHOST
    cluster_start
}

start_master()
{
    echo_title "Start the Pacemaker as MASTER"
    cluster_start
    cluster_wait_current_dc
}

setup_cluster()
{
    local pcs_opt
    echo_title "Configure clustering settings."
    local timeout=60
    while [ $timeout -gt 0 ]; do
        let timeout=$timeout-1
        if verbose_run pcs cluster cib $TMPDIR/cib-org.xml; then
            break
        fi
        sleep 1
    done
    if [ $timeout -eq 0 ]; then
        abort_setup "Failed to get cib"
    fi
    verbose_run cp $TMPDIR/cib-org.xml $TMPDIR/cib-new.xml
    pcs_opt="-f $TMPDIR/cib-new.xml"
    pcs_setup_stonith $pcs_opt
    pcs_remove_resources $pcs_opt
    pcs_setup_resources $pcs_opt
    pcs_setup_constraints $pcs_opt
    verbose_run pcs cluster cib-push $TMPDIR/cib-new.xml
    if $CLUSTER_CONFIGURED; then
        pcs_failcount_reset res_rabbitmq res_pgsql
    fi
}

pcs_setup_stonith()
{
    local pcs_opt=$@
    pcs $pcs_opt property set no-quorum-policy="ignore" stonith-enabled="false"
}

pcs_setup_resources()
{
    local pcs_opt=$@
    local RES_VIP=""
    local RES_KOMPIRA_JOBMNGRD=""
    local class=""
    local params_res_vip

    #
    # create webserver group resources
    #
    verbose_run pcs $pcs_opt resource create res_httpd ocf:heartbeat:apache \
        configfile="/etc/httpd/conf/httpd.conf" envfiles="/etc/sysconfig/httpd" port="80" statusurl="http://localhost/.login" \
        op start interval="0" timeout="60s" \
        op monitor interval="5s" timeout="20s" \
        op stop interval="0" timeout="60s"
    verbose_run pcs $pcs_opt resource create res_kompirad ocf:kompira:kompirad \
        op start interval="0" timeout="60s" \
        op monitor interval="5s" timeout="20s" \
        op stop interval="0" timeout="60s"
    if ! $NO_JOBMNGR_SETUP; then
        verbose_run pcs $pcs_opt resource create res_kompira_jobmngrd ocf:kompira:kompira_jobmngrd \
            op start interval="0" timeout="60s" \
            op monitor interval="5s" timeout="20s" \
            op stop interval="0" timeout="60s"
        RES_KOMPIRA_JOBMNGRD="res_kompira_jobmngrd"
    fi
    if ! $NO_CLUSTER_VIP; then
        # ip="$cluster_vip" nic="$ETH0" cidr_netmask="$cluster_netmask" \
        params_res_vip="ip=$cluster_vip cidr_netmask=$cluster_netmask"
        if [ -n "$cluster_nic" ]; then
            params_res_vip="$params_res_vip nic=$cluster_nic"
        fi
        verbose_run pcs $pcs_opt resource create res_vip ocf:heartbeat:IPaddr2 \
            $params_res_vip \
            op monitor interval="10s"
        RES_VIP="res_vip"
    fi
    verbose_run pcs $pcs_opt resource create res_lsyncd $PCS_STD_RESOURCE_CLASS:lsyncd \
        op start interval="0s" timeout="20s" \
        op stop interval="0s" timeout="20s" \
        op monitor interval="5s" timeout="20s"

    verbose_run pcs $pcs_opt resource create res_pgsql ocf:heartbeat:pgsql \
        pgctl="$(pgsql_cmddir)/pg_ctl" \
        psql="$(pgsql_cmddir)/psql" \
        pgdata="$PG_DATADIR" \
        tmpdir="$PG_BASEDIR/tmp" \
        rep_mode="async" \
        repuser="$PG_REPL_USER" \
        restore_command="test -f %p && test ! -f $PG_BASEDIR/pg_archive/%f && cp %p $PG_BASEDIR/pg_archive/%f" \
        primary_conninfo_opt="keepalives_idle=60 keepalives_interval=5 keepalives_count=5" \
        node_list="$primary_name $secondary_name" \
        master_ip="$HA_OTHERHOST" \
        op start   timeout="60s" interval="0s" on-fail="restart" \
        op monitor timeout="60s" interval="7s" on-fail="restart" \
        op monitor timeout="60s" interval="5s" on-fail="restart" role="Master" \
        op promote timeout="60s" interval="0s" on-fail="restart" \
        op demote  timeout="60s" interval="0s" on-fail="stop" \
        op stop    timeout="60s" interval="0s" on-fail="block" \
        op notify  timeout="60s" interval="0s"

    verbose_run pcs $pcs_opt resource create res_rabbitmq ocf:rabbitmq:rabbitmq-server \
        mnesia_base="/var/lib/rabbitmq/mnesia" nodename="" \
        op monitor interval="7s" timeout="20s" \
        op monitor interval="5s" timeout="20s" role="Master" \
        op start interval="0" timeout="600" \
        op stop interval="0" timeout="120"

    verbose_run pcs $pcs_opt resource group add webserver $RES_VIP res_httpd res_kompirad $RES_KOMPIRA_JOBMNGRD res_lsyncd
    verbose_run pcs $pcs_opt resource meta webserver target-role="Started"
    verbose_run pcs $pcs_opt resource master ms_pgsql res_pgsql \
        master-max=1 master-node-max=1 clone-max=2 clone-node-max=1 notify=true
    verbose_run pcs $pcs_opt resource master ms_rabbitmq res_rabbitmq \
        clone-max=2 notify=true
    verbose_run pcs $pcs_opt resource defaults resource-stickness="INFINITY" migration-threshold="1"
}

usage_exit()
{
    cat <<EOF
Usage: `basename $0` [OPTIONS] <cluster-vip>/<prefix>

  --cluster-name        Specify the cluster name (up to 15 characters).
  --primary             Setup for primary server.
  --secondary           Setup for secondary server.
  --master-mode         Setup with master mode.
  --slave-mode          Setup with slave mode.
  --without-vip         Setup without VIP.
  --without-jobmanager  Setup without job manager.
  --hostname-prefix=PREFIX_NAME
                        Specify prefix of hostname.
  --heartbeat-netaddr=NETWORK_ADDRESS/PREFIX
                        Specify network address for heartbeat.
  --manual-heartbeat    Specify heartbeat network manually.
  --heartbeat-device=DEVICE
                        Specify network interface for heartbeat.
  --heartbeat-primary=NETWORK_ADDRESS
                        Manual address setting for heartbeat-primary.
  --heartbeat-secondary=NETWORK_ADDRESS
                        Manual address setting for heartbeat-secondary.
  --dry-run             Check the parameters without actually installing.
  --help                Show this message.

EOF
    exit 1
}

parse_options()
{
    OPTIONS=`getopt -q -o '' -l cluster-name:,primary,secondary,master-mode,slave-mode,without-vip,without-jobmanager,manual-heartbeat,hostname-prefix:,heartbeat-netaddr:,heartbeat-unicast,heartbeat-primary:,heartbeat-secondary:,heartbeat-device:,cluster-device:,proxy:,noproxy:,manual,offline,dry-run,help -- "$@"`
    [ $? != 0 ] && usage_exit

    local slave_mode
    eval set -- "$OPTIONS"
    while true
    do
        arg="$1"
        case "$arg" in
            --cluster-name)
                HA_CLUSTER_NAME="$2"
                shift ;;
            --primary)
                SECONDARY=false
                SLAVE_MODE=false ;;
            --secondary)
                SECONDARY=true
                SLAVE_MODE=true  ;;
            --master-mode)
                slave_mode=false ;;
            --slave-mode)
                slave_mode=true ;;
            --without-vip)
                NO_CLUSTER_VIP=true ;;
            --without-jobmanager)
                NO_JOBMNGR_SETUP=true ;;
            --manual-heartbeat)
                MANUAL_HEARTBEAT=true ;;
            --skip-hostname-setup)
                SKIP_HOSTNAME=true ;;
            --hostname-prefix)
                HA_HOSTNAME_PREFIX="$2"
                shift ;;
            --heartbeat-netaddr)
                HEARTBEAT_NETCIDR="$2"
                shift ;;
            --manual)
                MANUAL_MODE=true ;;
            --heartbeat-device)
                HEARTBEAT_DEVICE="$2"
                shift ;;
            --heartbeat-unicast)
                HEARTBEAT_UNICAST=true ;;
            --heartbeat-primary)
                PRIMARY_IP="$2"
                shift ;;
            --heartbeat-secondary)
                SECONDARY_IP="$2"
                shift ;;
            --cluster-device)
                CLUSTER_DEVICE="$2"
                shift ;;
            --proxy)
                PROXY_URL=$(normalize_proxy_url $2)
                shift ;;
            --noproxy)
                NO_PROXY=$2
                shift ;;
            --offline)
                OFFLINE_MODE=true
                SETUP_TYPE="$SETUP_TYPE (offline-mode)" ;;
            --dry-run)
                DRY_RUN=true ;;
            --) shift; break ;; # 引数は無視
            *) usage_exit ;;
        esac
        shift
    done

    if [ -n "$slave_mode" ]; then
        SLAVE_MODE=$slave_mode
    fi

    if [ -z "$HA_HOSTNAME_PREFIX" ]; then
        echo "ERROR: hostname-prefix must be non-empty string."
        echo
        usage_exit
    fi

    if $MANUAL_HEARTBEAT; then
        if [ -z "$PRIMARY_IP" -o -z "$SECONDARY_IP" ]; then
            echo "ERROR: option --heartbeat-primary and --heartbeat-secondary are required when manual-heartbeat."
            echo
            usage_exit
        fi
        if ! check_ipaddr $PRIMARY_IP; then
            echo "ERROR: invalid heartbeat-primary: $PRIMARY_IP"
            echo
            usage_exit
        fi
        if ! check_ipaddr $SECONDARY_IP; then
            echo "ERROR: invalid heartbeat-secondary: $SECONDARY_IP"
            echo
            usage_exit
        fi
    else
        if [ -z "$HEARTBEAT_DEVICE" ]; then
            echo "ERROR: option --heartbeat-device is required"
            echo
            usage_exit
        fi
        if ! check_cidr $HEARTBEAT_NETCIDR; then
            echo "ERROR: invalid heartbeat-netaddr: $HEARTBEAT_NETCIDR"
            echo
            usage_exit
        fi
    fi

    if ! $NO_CLUSTER_VIP; then
        if ! $SLAVE_MODE && ! check_cidr "$1"; then
            echo "ERROR: invalid <cluster-vip>/<prefix>."
            echo
            usage_exit
        else
            CLUSTER_VIP="$1"
        fi
    fi
}


################################################################

parse_options "$@"
exec > >(tee $SETUP_LOG) 2>&1
create_tmpdir
start_setup
if ! $MANUAL_MODE; then
    set_params
fi
show_params
check_params
check_root
check_environ
if $DRY_RUN ; then
    exit_setup
fi

install_cluster_packages
stop_pacemaker
setup_network
setup_apache
setup_kompira
setup_rsync
setup_pgsql
setup_rabbitmq
setup_pacemaker
if $SLAVE_MODE; then
    start_slave
else
    start_master
    setup_cluster
fi
exit_setup
